#!/usr/bin/env python

"""
Change the speed of an audio file without changing its pitch (with GStreamer).
"""

import argparse
import os
import sys
import gi
gi.require_version('Gst', '1.0')

# pylint: disable=wrong-import-position
from gi.repository import Gst, GObject
Gst.init(sys.argv)
import audiotsm.gstreamer.wsola  # pylint: disable=unused-import
# pylint: enable=wrong-import-position


class Pipeline(Gst.Pipeline):
    """A gstreamer pipeline that changes the speed of an audio file, write the
    output to another file, and quit the main gobject loop with this is done.

    :param tsm_description: the description used to create the tsm element of
        the pipeline. This will be used as input of the
        :func:`Gst.parse_launch` function.
    """
    def __init__(self, tsm_description="audiotsm-wsola"):
        super().__init__()

        self._speed = 0

        # Create the playbin, that will handle the decoding of the audio files
        self.playbin = Gst.ElementFactory.make('playbin', 'playbin')
        self.add(self.playbin)

        # Create the audiotsm bin, that will handle the TSM
        audiotsmbin = Gst.Bin('audiotsm')

        # Create the elements of the audiotsm bin, add them, and link them
        self.tsm = Gst.parse_launch(tsm_description)
        converter = Gst.ElementFactory.make('audioconvert', 'converter')
        encoder = Gst.ElementFactory.make('wavenc', 'encoder')
        self.sink = Gst.ElementFactory.make('filesink', 'sink')

        audiotsmbin.add(self.tsm)
        audiotsmbin.add(converter)
        audiotsmbin.add(encoder)
        audiotsmbin.add(self.sink)

        self.tsm.link(converter)
        converter.link(encoder)
        encoder.link(self.sink)

        # Add the sink pad of the TSM plugin to the audiotsm bin.
        self.tsm_sink_pad = Gst.GhostPad.new(
            'sink', self.tsm.get_static_pad('sink'))
        audiotsmbin.add_pad(self.tsm_sink_pad)

        # And link it to the playbin
        self.playbin.set_property("audio-sink", audiotsmbin)

        bus = self.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", self._on_error)
        bus.connect("message::eos", self._on_eos)

    def _on_error(self, _, message):
        """Called when there is an error during the playback."""
        # pylint: disable=no-self-use
        err, debug = message.parse_error()
        print("Error: %s" % err, debug)
        sys.exit()

    def _on_eos(self, _1, _2):
        """Called when the end of the audio file is reached."""
        # pylint: disable=no-self-use
        sys.exit()

    def set_speed(self, speed):
        """Set the speed ratio."""
        self._speed = speed

    def save(self, path):
        """Save the output of the TSM procedure to path, then quit the GObject
        loop."""
        self.sink.set_property('location', path)
        self.set_state(Gst.State.PAUSED)
        self.get_state(Gst.CLOCK_TIME_NONE)

        self.playbin.seek(self._speed, Gst.Format.BYTES, Gst.SeekFlags.FLUSH,
                          Gst.SeekType.SET, 0, Gst.SeekType.NONE, -1)

        self.set_state(Gst.State.PLAYING)

    def set_src_uri(self, uri):
        """Set the uri of the source audio file."""
        self.playbin.set_property("uri", uri)


def main():
    """Change the speed of an audio file without changing its pitch."""

    parser = argparse.ArgumentParser(description=(
        "Change the speed of an audio file without changing its pitch."))
    parser.add_argument(
        '-s', '--speed', metavar="S", type=float, default=1.,
        help="Set the speed ratio (e.g 0.5 to play at half speed)")
    parser.add_argument(
        'input_filename', metavar='INPUT_FILENAME', type=str,
        help="The audio input file")
    parser.add_argument(
        'output_filename', metavar='OUTPUT_FILENAME', type=str,
        help="The audio output file")

    args = parser.parse_args()

    if not os.path.isfile(args.input_filename):
        parser.error(
            'The input file "{}" does not exist.'.format(args.input_filename))

    pipeline = Pipeline()
    pipeline.set_speed(args.speed)
    pipeline.set_src_uri('file:///' + os.path.realpath(args.input_filename))
    pipeline.save(args.output_filename)

    loop = GObject.MainLoop()
    loop.run()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.gstreamer` module implements three audio filters allowing to
use the TSM procedures with gstreamer:

- ``audiotsm-ola``, defined in the :mod:`audiotsm.gstreamer.ola` module;
- ``audiotsm-wsola``, defined in the :mod:`audiotsm.gstreamer.wsola` module;
- ``audiotsm-phase-vocoder``, defined in the
  :mod:`audiotsm.gstreamer.phasevocoder` module.

.. note::

    If you are unsure which filter to choose, using ``audiotsm-phase-vocoder``
    should give good results in most cases. You can listen to the output of the
    different procedures on various audio files and at various speeds on the
    `examples page`_.

In order to use these audio filters, you should first import the module
corresponding to the TSM procedure you want to use, for example::

    import audiotsm.gstreamer.phasevocoder

Then, you should create the audio filter with ``Gst.ElementFactory.make``, as
follow::

    tsm = Gst.ElementFactory.make("audiotsm-phase-vocoder")

You should then create a gstreamer pipeline using the audio filter you created.
See ``examples/audiotsmcli_gst.py`` for an example of pipeline.

The audio filters work in the same manner as the ``scaletempo`` gstreamer
plugin. You can change the playback rate by sending a seek event to the
pipeline::

    speed = 0.5
    pipeline.seek(speed, Gst.Format.BYTES, Gst.SeekFlags.FLUSH,
                  Gst.SeekType.NONE, -1, Gst.SeekType.NONE, -1)

The other parameters of the TSM procedure are available as properties, as
documented for each of the procedures below.

.. _examples page: https://muges.github.io/audiotsm/
"""



# -*- coding: utf-8 -*-

"""
Tests the TSM procedures on real data.
"""

import os
import shutil
import pytest

from audiotsm import ola, wsola, phasevocoder, PhaseLocking
from audiotsm.io.wav import WavReader, WavWriter
from audiotsm.io.array import ArrayWriter

EXAMPLES_DIR = os.path.join("build", "ghpages", "examples")


def create_tsm(name, channels, speed):
    """Create a TSM object given the method name and its parameters."""
    if name == "ola":
        return ola(channels, speed)
    if name == "wsola":
        return wsola(channels, speed)
    if name == "phasevocoder":
        return phasevocoder(channels, speed, phase_locking=PhaseLocking.NONE)
    if name == "phasevocoder_identity":
        return phasevocoder(channels, speed,
                            phase_locking=PhaseLocking.IDENTITY)

    raise ValueError("unknown TSM method: {}".format(name))


def test_data(data_file, speed, tsm_name, save):
    """Test the TSM procedures on real data."""
    reader = None
    writer = None

    try:
        # Create the reader
        reader = WavReader(data_file)

        # Create the writer
        if save:
            # pylint: disable=no-member
            rel_path = os.path.relpath(data_file, pytest.DATA_DIR)
            # pylint: enable=no-member

            # Copy original file to "orig" directory
            orig_file = os.path.join(EXAMPLES_DIR, "orig", rel_path)
            orig_dir = os.path.dirname(orig_file)
            if not os.path.isdir(orig_dir):
                os.makedirs(orig_dir)
            if not os.path.isfile(orig_file):
                shutil.copy2(data_file, orig_file)

            # Generate output file path
            speed_dir = "speed-{:.2f}".format(speed)
            name = os.path.splitext(rel_path)[0]
            output_name = "{}_{}.wav".format(name, tsm_name)
            output_file = os.path.join(EXAMPLES_DIR, speed_dir, output_name)
            output_dir = os.path.dirname(output_file)
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)

            writer = WavWriter(output_file, reader.channels, reader.samplerate)
        else:
            writer = ArrayWriter(reader.channels)

        # Create and run the TSM
        tsm = create_tsm(tsm_name, reader.channels, speed)
        tsm.run(reader, writer)

    finally:
        # Close files
        if reader:
            reader.close()
        if save and writer:
            writer.close()



# -*- coding: utf-8 -*-

"""
The :mod:`~audiotsm.gstreamer.base` module provides a base class for gstreamer
plugin using :class:`~audiotsm.base.tsm.TSM` objects.
"""

import numpy as np
import gi
gi.require_version('Gst', '1.0')
gi.require_version('GstAudio', '1.0')

# pylint: disable=wrong-import-position
from gi.repository import GObject, GLib, Gst, GstAudio
from audiotsm import __version__
from audiotsm.io.array import ArrayReader, ArrayWriter
from gstbasetransform import BaseTransform
# pylint: enable=wrong-import-position

CAPS = Gst.Caps.from_string(
    "audio/x-raw,format=S16LE,layout=interleaved")


def audioformatinfo_to_dtype(info):
    """Return the data type corresponding to a ``GstAudio.AudioFormatInfo``
    object.

    :param info: a ``GstAudio.AudioFormatInfo``.
    :returns: the corresponding data type, to be used in :mod:`numpy`
        functions.
    """
    endianness = '<' if info.endianness == GLib.LITTLE_ENDIAN else '>'

    if info.flags & GstAudio.AudioFormatFlags.INTEGER:
        if info.flags & GstAudio.AudioFormatFlags.SIGNED:
            _type = 'i'
        else:
            _type = 'u'
    elif info.flags & GstAudio.AudioFormatFlags.FLOAT:
        _type = 'f'
    else:
        raise ValueError(
            'unsupported audio format flags: {}'.format(info.flags))

    samplewidth = info.width // 8  # in bytes

    return '{}{}{}'.format(endianness, _type, samplewidth)


class GstTSM(BaseTransform):
    """Gstreamer TSM plugin.

    Subclasses should implement the :func:`~GstTSM.create_tsm` method and
    provide two class attributes:

    - ``__gstmetadata__ = (longname, classification, description, author)``.
      See the documentation of the gst_element_class_set_metadata_ function for
      more details.
    - ``plugin_name``, the name of the plugin.

    Calling the :func:`~GstTSM.register` class method on a subclass will
    register it, enabling you to instantiate an audio filter with
    ``Gst.ElementFactory.make(plugin_name)``.

    .. _gst_element_class_set_metadata:
        https://gstreamer.freedesktop.org/data/doc/gstreamer/head/gstreamer/html/GstElement.html#gst-element-class-set-metadata
    """  # noqa: E501
    # pylint: disable=no-member

    __gsttemplates__ = (Gst.PadTemplate.new("src",
                                            Gst.PadDirection.SRC,
                                            Gst.PadPresence.ALWAYS,
                                            CAPS),
                        Gst.PadTemplate.new("sink",
                                            Gst.PadDirection.SINK,
                                            Gst.PadPresence.ALWAYS,
                                            CAPS))

    def __init__(self):
        super().__init__()

        self._channels = 0
        self._samplerate = 0
        self._bps = 0  # bytes per sample
        self._dtype = ''
        self._audioformatinfo = None

        self._tsm = None
        self._position = 0

    @classmethod
    def plugin_init(cls, plugin):
        """Initialize the plugin."""
        plugin_type = GObject.type_register(cls)
        Gst.Element.register(plugin, cls.plugin_name, 0, plugin_type)
        return True

    @classmethod
    def register(cls):
        """Register the plugin.

        Register the plugin to make it possible to instantiate it with
        ``Gst.ElementFactory.make``."""
        Gst.Plugin.register_static(
            Gst.VERSION_MAJOR, Gst.VERSION_MINOR, cls.plugin_name,
            cls.get_metadata('description'), cls.plugin_init, __version__,
            'MIT/X11', 'audiotsm', 'audiotsm', '')

    def _gstbuffer_to_ndarray(self, gstbuffer):
        """Return the data contained in ``gstbuffer`` as a
        :class:`numpy.ndarray`.

        :param gstbuffer: a :class:`Gst.Buffer`.
        """
        _, mapinfo = gstbuffer.map(Gst.MapFlags.READ)
        data = mapinfo.data
        gstbuffer.unmap(mapinfo)

        data = np.frombuffer(data, self._dtype).astype(np.float32) / 32767
        data = data.reshape((-1, self._channels)).T

        return data

    def _ndarray_to_gstbuffer(self, gstbuffer, data):
        """Write the ``data`` to ``gstbuffer``.

        This method is a bit hack-ish. It would be better to set the size of
        the :class:`Gst.Buffer` in advance with the
        :func:`GstBase.BaseTransform.do_transform_size` virtual method, but it
        is not possible to do so with pygst.

        :param gstbuffer: a :class:`Gst.Buffer`.
        :param data: a :class:`numpy.ndarray`.
        """
        length = data.shape[1]

        if length <= 0:
            gstbuffer.set_size(0)
            gstbuffer.duration = 0
            return

        np.clip(data, -1, 1, out=data)
        data = (data.T.reshape((-1,)) * 32767).astype(self._dtype).tobytes()
        size = len(data)

        # Copy as many bytes as possible to the buffer directly
        n = gstbuffer.fill(0, data)

        if n < size:
            Gst.warning('the output buffer is too small, allocating memory')

            # Allocate memory for the rest of the data
            # This may add noise to the audio signal
            data = data[n:]
            mem = Gst.Memory.new_wrapped(0, data, len(data), 0, None, None)
            gstbuffer.append_memory(mem)

        gstbuffer.set_size(size)

    def do_sink_event(self, event):
        """Sink pad event handler."""
        if event.type == Gst.EventType.CAPS:
            # CAPS event, used to negotiate the format with the "upstream"
            # gstreamer element.
            caps = event.parse_caps()
            structure = caps.get_structure(0)

            # Ensure that the layout is interleaved
            layout = structure.get_string('layout')
            if layout != 'interleaved':
                # Returns False if we were unable to agree on a format.
                return False

            # Get number of channels
            success, self._channels = structure.get_int('channels')
            if not success:
                # Returns False if we were unable to agree on a format.
                return False

            # Get samplerate
            success, self._samplerate = structure.get_int('rate')
            if not success:
                # Returns False if we were unable to agree on a format.
                return False

            # Get and parse samples format
            samples_format = structure.get_string('format')
            if samples_format is None:
                # Returns False if we were unable to agree on a format.
                return False

            self._audioformatinfo = GstAudio.AudioFormat.get_info(
                GstAudio.AudioFormat.from_string(samples_format)
            )
            self._dtype = audioformatinfo_to_dtype(self._audioformatinfo)

            self._bps = self._channels * self._audioformatinfo.width // 8

            # Create the TSM object
            self._tsm = self.create_tsm(self._channels)

        if event.type == Gst.EventType.SEGMENT:
            segment = event.parse_segment()

            self._tsm.set_speed(segment.rate)
            self._position = segment.position

            segment.applied_rate = segment.rate
            segment.rate = 1.0
            event = Gst.Event.new_segment(segment)
            self.srcpad.push_event(event)

        if event.type == Gst.EventType.EOS:
            # Flush the TSM object at the end of the stream
            writer = ArrayWriter(self._channels)
            self._tsm.flush_to(writer)

            # Write the output to a Gst.Buffer
            out_buffer = Gst.Buffer.new()
            self._ndarray_to_gstbuffer(out_buffer, writer.data)

            out_buffer.pts = self._position
            self._position += out_buffer.duration

            # Send the buffer downstream
            self.srcpad.push(out_buffer)

        # Propagate the event downstream
        return self.srcpad.push_event(event)

    def do_transform(self, in_buffer, out_buffer):
        """Run the data of ``in_buffer`` through the
        :class:`~audiotsm.base.tsm.TSM` object and write the output to
        ``out_buffer``.

        :param in_buffer: a ``Gst.Buffer`` containing the input data.
        :param out_buffer: a ``Gst.Buffer`` where the output data will be
            written.
        """
        # There is a bug that increases the refcount of out_buffer, making it
        # non writable (https://bugzilla.gnome.org/show_bug.cgi?id=727702#c4).
        # Set the refcount to 1 to fix this.
        refcount = out_buffer.mini_object.refcount
        out_buffer.mini_object.refcount = 1

        # Set the position of the output buffer
        out_buffer.pts = self._position

        # Run the TSM procedure
        reader = ArrayReader(self._gstbuffer_to_ndarray(in_buffer))
        writer = ArrayWriter(self._channels)

        self._tsm.run(reader, writer, flush=False)

        self._ndarray_to_gstbuffer(out_buffer, writer.data)
        out_buffer.duration = (
            (out_buffer.get_size() * Gst.SECOND) //
            (self._bps * self._samplerate)
        )
        self._position += out_buffer.duration

        # Reset the refcount
        out_buffer.mini_object.refcount = refcount

        return Gst.FlowReturn.OK

    def do_transform_size(self, direction, caps, size, othercaps):
        """Returns the size of the output buffer given the size of the input
        buffer."""
        input_length = size // self._bps
        output_length = self._tsm.get_max_output_length(input_length)
        output_size = output_length * self._bps
        return True, output_size

    def create_tsm(self, channels):
        """Returns the :class:`~audiotsm.base.tsm.TSM` object used by the audio
        filter."""
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Generate example page for github pages.
"""

import os
import re
from distutils.dir_util import copy_tree
from docutils.core import publish_parts
from jinja2 import Environment, FileSystemLoader, Markup
import sass

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT = os.path.join(SCRIPT_DIR, "..", "..")

ASSETS_DIR = os.path.join(SCRIPT_DIR, "assets")
TEMPLATES_DIR = os.path.join(SCRIPT_DIR, "templates")
CREDITS_PATH = os.path.join(ROOT, "tests", "integration", "data", "README.rst")
DESTINATION_DIR = os.path.join(ROOT, "build", "ghpages")
EXAMPLES_DIR = os.path.join(DESTINATION_DIR, "examples")

CSS_FILE = os.path.join(SCRIPT_DIR, "scss", "audiotsm.scss")

ENVIRONMENT = Environment(
    loader=FileSystemLoader(TEMPLATES_DIR),
    autoescape=True, trim_blocks=False)

SPEED_DIR_RE = re.compile(r'speed-([\d\.]+)')
FILE_RE = re.compile(r'([^_]+)_(.+)\.wav')

METHOD_NAME = {
    'ola': 'OLA',
    'wsola': 'WSOLA',
    'phasevocoder': Markup('Phase<br/>Vocoder'),
    'phasevocoder_identity': Markup(
        'Phase Vocoder<br/>(identity)'),
}
METHOD_SORT = ['ola', 'wsola', 'phasevocoder', 'phasevocoder_identity']


def get_examples():
    speeds = set()
    methods = set()
    files = {}

    for speed_dir in os.listdir(EXAMPLES_DIR):
        match = SPEED_DIR_RE.match(speed_dir)
        if match:
            speed = match.group(1)
            speeds.add(speed)
            files[speed] = {}
        elif speed_dir == 'orig':
            speed = 'orig'
            files[speed] = {}
        else:
            print('invalid speed directory "{}"'.format(speed_dir))
            continue

        speed_dir = os.path.join(EXAMPLES_DIR, speed_dir)
        for root, _, filenames in os.walk(speed_dir):
            for filename in filenames:
                fullpath = os.path.join(root, filename)
                path = os.path.relpath(fullpath, DESTINATION_DIR)
                name = os.path.relpath(fullpath, speed_dir)

                if speed == 'orig':
                    name = os.path.splitext(name)[0]
                    files[speed][name] = path
                else:
                    match = FILE_RE.match(name)
                    if match:
                        name = match.group(1)
                        method = match.group(2)
                        methods.add(method)
                        if name not in files[speed]:
                            files[speed][name] = {}
                        files[speed][name][method] = path
                    else:
                        print('invalid example file "{}"'.format(filename))

    return {
        'speeds': sorted(speeds, key=float),
        'methods': [
            (method, METHOD_NAME[method])
            for method in sorted(methods, key=METHOD_SORT.index)
        ],
        'files': files
    }


def generate_css():
    """Generate css stylesheet."""
    css = sass.compile(filename=CSS_FILE)

    css_dir = os.path.join(DESTINATION_DIR, "css")
    if not os.path.isdir(css_dir):
        os.makedirs(css_dir)

    filename = os.path.join(css_dir, "audiotsm.css")
    with open(filename, 'w') as fileobj:
        fileobj.write(css)


def generate_index():
    """Generate index.html."""
    context = get_examples()

    with open(CREDITS_PATH) as fileobj:
        context['credits'] = Markup(
            publish_parts(fileobj.read(), writer_name='html')['fragment'])

    filename = os.path.join(DESTINATION_DIR, 'index.html')
    with open(filename, 'w') as fileobj:
        template = ENVIRONMENT.get_template('index.html')
        html = template.render(context)
        fileobj.write(html)


def main():
    """Generate github pages."""
    # Copy assets
    copy_tree(ASSETS_DIR, DESTINATION_DIR)

    generate_css()
    generate_index()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm` module provides several time-scale modification procedures:

- :func:`~audiotsm.ola` (Overlap-Add);
- :func:`~audiotsm.wsola` (Waveform Similarity-based Overlap-Add);
- :func:`~audiotsm.phasevocoder` (Phase Vocoder).

The OLA procedure should only be used on percussive audio signals. The WSOLA
and the Phase Vocoder procedures are improvements of the OLA procedure, and
should both give good results in most cases.

.. note::

    If you are unsure which procedure and parameters to choose, using
    :func:`~audiotsm.phasevocoder` with the default parameters should give good
    results in most cases. You can listen to the output of the different
    procedures on various audio files and at various speeds on the `examples
    page`_.

Each of the function of this module returns a :class:`~audiotsm.base.tsm.TSM`
object which implements a time-scale modification procedure.

.. autofunction:: audiotsm.ola
.. autofunction:: audiotsm.wsola
.. autofunction:: audiotsm.phasevocoder
.. autoclass:: audiotsm.PhaseLocking
    :members:

.. _examples page: https://muges.github.io/audiotsm/
"""

__version__ = "0.1.2"


from .ola import ola
from .wsola import wsola
from .phasevocoder import phasevocoder, PhaseLocking



# -*- coding: utf-8 -*-

"""
Pytest configuration
"""

import fnmatch
import os
import pytest


def pytest_namespace():
    """Set pytest global variables."""
    return {'DATA_DIR': os.path.join("tests", "integration", "data")}


def pytest_addoption(parser):
    """Add command line options to the pytest command."""
    parser.addoption(
        "--data-all", action="store_true", help="run on all data files")
    parser.addoption(
        "--data-save", action="store_true",
        help="save the output of the data tests to build/examples")


def get_data_files():
    """Recursively look for wav files in the DATA_DIR directory, and returns
    their paths."""
    # pylint: disable=no-member
    data_files = []
    for root, _, filenames in os.walk(pytest.DATA_DIR):
        for filename in fnmatch.filter(filenames, '*.wav'):
            data_files.append(os.path.join(root, filename))

    return data_files


def pytest_generate_tests(metafunc):
    """Generate tests for test_data.py."""
    if 'speed' in metafunc.fixturenames:
        metafunc.parametrize('speed', [1 / 1.2, 1 / 1.8, 2])

    if 'tsm_name' in metafunc.fixturenames:
        metafunc.parametrize('tsm_name', [
            "ola", "wsola", "phasevocoder", "phasevocoder_identity"])

    if 'save' in metafunc.fixturenames:
        metafunc.parametrize("save", [metafunc.config.getoption('data_save')])

    if 'data_file' in metafunc.fixturenames:
        data_files = get_data_files()
        if not metafunc.config.getoption('data_all'):
            # Only test with two files
            data_files = data_files[:2]
        metafunc.parametrize("data_file", data_files)



#!/usr/bin/env python

"""
A simple gtk audio player using audiotsm with gstreamer.
"""

import os
import sys
import numpy as np
import gi
gi.require_version('Gst', '1.0')
gi.require_version('Gtk', '3.0')

# pylint: disable=wrong-import-position
from gi.repository import Gst, GObject, Gtk
Gst.init(sys.argv)
import audiotsm.gstreamer.phasevocoder  # pylint: disable=unused-import
# pylint: enable=wrong-import-position


def get_icon_name(icon_names):
    """Return the first icon name that is available in the current GTK icon
    theme."""
    theme = Gtk.IconTheme.get_default()

    for icon_name in icon_names:
        if theme.has_icon(icon_name):
            return icon_name

    return ""


def format_time(time):
    """Format a time as "minutes:seconds"."""
    minutes, seconds = divmod(time // Gst.SECOND, 60)
    return "{}:{:02d}".format(minutes, seconds)


class Player(Gst.Pipeline):
    """A gstreamer pipeline that provides an easy-to-use player with a
    changeable playing speed.

    :param tsm_description: the description used to create the tsm element of
        the pipeline. This will be used as input of the
        :func:`Gst.parse_launch` function.

    Signals
    ~~~~~~~

    ``position-changed(position, duration)``
        Emitted at regular time interval during the playback.
    ``state-changed(state)``
        Emitted when the state of the player changes.
    """
    __gsignals__ = {
        'position-changed':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_LONG,
                                              GObject.TYPE_LONG)),
        'state-changed':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_INT,))
    }

    def __init__(self, tsm_description="audiotsm-phase-vocoder"):
        super().__init__()

        self._speed = 1.0
        self._speed_set = False

        # Create the playbin, that will handle the decoding of the audio files
        self.playbin = Gst.ElementFactory.make('playbin', 'playbin')
        self.add(self.playbin)

        # Create the audiotsm bin, that will handle the TSM
        audiotsmbin = Gst.Bin('audiotsm')

        # Create the elements of the audiotsm bin, add them, and link them
        self.tsm = Gst.parse_launch(tsm_description)
        converter = Gst.ElementFactory.make('audioconvert', 'converter')
        sink = Gst.ElementFactory.make('autoaudiosink', 'sink')

        audiotsmbin.add(self.tsm)
        audiotsmbin.add(converter)
        audiotsmbin.add(sink)

        self.tsm.link(converter)
        converter.link(sink)

        # Add the sink pad of the TSM plugin to the audiotsm bin.
        self.tsm_sink_pad = Gst.GhostPad.new(
            'sink', self.tsm.get_static_pad('sink'))
        audiotsmbin.add_pad(self.tsm_sink_pad)

        # And link it to the playbin
        self.playbin.set_property("audio-sink", audiotsmbin)

        bus = self.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", self._on_error)
        bus.connect("message::eos", self._on_eos)
        bus.connect("message::state-changed", self._on_state_changed)

        # The timer that will emit the position-changed signal
        self.position_timer = None

    def do_state_changed(self, *args):
        # pylint: disable=missing-docstring
        # Override the do_state_changed method that is automatically called
        # when the state-changed signal is emitted, and raises an exception.
        pass

    def _emit_position(self):
        """Emit a position-changed signal. Called every 200ms during
        playback."""
        position = self.get_position()
        duration = self.get_duration()
        if duration != 0:
            self.emit('position-changed', position, duration)

        return True

    def get_duration(self):
        """Return the duration of the file currently played, or 0 if it is
        unknown.

        This is not affected by the speed ratio."""
        success, duration = self.query_duration(Gst.Format.TIME)
        if success:
            return duration

        return 0

    def get_position(self):
        """Return the current playback position of the file, or 0 if it is
        unknown.

        This is not affected by the speed ratio."""
        success, position = self.playbin.query_position(Gst.Format.TIME)
        if success:
            return position

        return 0

    def _on_error(self, bus, message):
        """Called when there is an error during the playback."""
        self.set_state(Gst.State.NULL)
        err, debug = message.parse_error()
        print("Error: %s" % err, debug)

    def _on_eos(self, bus, message):
        """Called when the end of the audio file is reached."""
        duration = self.get_duration()
        self.emit('position-changed', duration, duration)
        self.set_state(Gst.State.READY)

    def _on_state_changed(self, bus, message):
        """Called when the state of an element of the pipeline changes."""
        state = message.parse_state_changed()[1]

        if message.src != self.playbin:
            # Ignore messages that do not come from the playbin
            return

        if state == Gst.State.READY:
            self._speed_set = False

        if state == Gst.State.PAUSED:
            self._emit_position()

        if state == Gst.State.PLAYING:
            if not self._speed_set:
                self._speed_set = True
                self.playbin.seek(
                    self._speed, Gst.Format.TIME, Gst.SeekFlags.FLUSH,
                    Gst.SeekType.SET, 0, Gst.SeekType.NONE, -1)

            if self.position_timer is None:
                self._emit_position()
                self.position_timer = (
                    GObject.timeout_add(100, self._emit_position)
                )
        else:
            if self.position_timer is not None:
                GObject.source_remove(self.position_timer)
                self.position_timer = None

        self.emit('state-changed', state)

    def seek(self, position):
        """Send a seek event to the playbin."""
        self.playbin.seek(self._speed, Gst.Format.TIME, Gst.SeekFlags.FLUSH,
                          Gst.SeekType.SET, position, Gst.SeekType.NONE, -1)

    def set_speed(self, speed):
        """Set the speed ratio."""
        if speed != self._speed:
            self._speed = speed

            position = self.get_position()
            self.playbin.seek(
                self._speed, Gst.Format.TIME, Gst.SeekFlags.FLUSH,
                Gst.SeekType.SET, position, Gst.SeekType.NONE, -1)

    def set_uri(self, uri):
        """Set the uri of the audio file."""
        self.set_state(Gst.State.NULL)
        self.playbin.set_property("uri", uri)
        self.set_state(Gst.State.PAUSED)

    def toggle(self):
        """Toggle between play and pause."""
        success, state, _ = self.get_state(0)
        if success != Gst.StateChangeReturn.SUCCESS:
            return

        if state == Gst.State.PLAYING:
            self.set_state(Gst.State.PAUSED)
        else:
            self.set_state(Gst.State.PLAYING)


class MainWindow(Gtk.Window):
    """The main window of audiotsmgtk."""
    def __init__(self):
        super().__init__(Gtk.WindowType.TOPLEVEL)
        self.set_title('audiotsmgtk')
        self.set_default_size(700, -1)
        self.connect('destroy', Gtk.main_quit)

        self._seeking = False

        # Headerbar
        headerbar = Gtk.HeaderBar()
        headerbar.set_title('audiotsmgtk')
        headerbar.set_show_close_button(True)
        self.set_titlebar(headerbar)

        # Open button
        icon_name = get_icon_name([
            'document-open-symbolic', 'document-open'])
        open_button = Gtk.Button.new_from_icon_name(
            icon_name, Gtk.IconSize.BUTTON)
        open_button.connect('clicked', self.on_open)
        headerbar.pack_start(open_button)

        # Play/pause button
        icon_name = get_icon_name([
            'media-playback-start-symbolic', 'media-playback-start'])
        self.toggle_button = Gtk.Button.new_from_icon_name(
            icon_name, Gtk.IconSize.BUTTON)
        self.toggle_button.connect('clicked', self.on_toggle)
        self.toggle_button.set_relief(Gtk.ReliefStyle.NONE)

        # Position and duration labels
        self.position_label = Gtk.Label("0:00")
        self.duration_label = Gtk.Label("0:00")

        # Seek bar
        self.seek_bar = Gtk.HScale.new_with_range(0, 100, 1)
        self.seek_bar.set_draw_value(False)
        self.seek_bar.connect('button-press-event', self.on_seeking_start)
        self.seek_bar.connect('button-release-event', self.on_seeking_end)

        # Speed
        speed_label = Gtk.Label.new("Speed:")
        speed_scale = Gtk.HScale.new_with_range(0, 100, 1)
        speed_scale.set_draw_value(False)
        for i in range(9):
            # Divide the speed scale in eight sections. Going from one marker
            # to the next will multiply the speed by two.
            speed_scale.add_mark(i * 12.5, Gtk.PositionType.BOTTOM, None)
        self.speed_label = Gtk.Label()
        speed_scale.connect('change-value', self.on_speed_scale_changed)
        speed_scale.connect('button-release-event',
                            self.on_speed_scale_released)
        speed_scale.set_value(50)
        self.speed_label.set_text('100.0 %')

        # Layout
        self.set_border_width(10)
        vbox = Gtk.VBox.new(False, 10)
        player_hbox = Gtk.HBox.new(False, 10)
        speed_hbox = Gtk.HBox.new(False, 10)

        player_hbox.pack_start(self.toggle_button, False, False, 0)
        player_hbox.pack_start(self.position_label, False, False, 0)
        player_hbox.pack_start(self.seek_bar, True, True, 0)
        player_hbox.pack_start(self.duration_label, False, False, 0)

        speed_hbox.pack_start(speed_label, False, False, 0)
        speed_hbox.pack_start(speed_scale, True, True, 0)
        speed_hbox.pack_start(self.speed_label, False, False, 0)

        vbox.pack_start(player_hbox, False, False, 0)
        vbox.pack_start(speed_hbox, False, False, 0)
        self.add(vbox)

        # Player
        self.player = Player()
        self.player.connect('position-changed', self.on_position_changed)
        self.player.connect('state-changed', self.on_state_changed)

    def open_file(self, path):
        """Open an audio file."""
        name = os.path.basename(path)
        self.set_title(name)

        uri = 'file://' + os.path.realpath(path)
        self.player.set_uri(uri)

    def on_open(self, button):
        """Called when the open button is clicked."""
        dialog = Gtk.FileChooserDialog(
            'Open audio file', self, Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            path = dialog.get_filename()
            self.open_file(path)

        dialog.destroy()

    def on_position_changed(self, player, position, duration):
        """Show the current playback position."""
        self.position_label.set_text(format_time(position))
        self.duration_label.set_text(format_time(duration))

        self.seek_bar.set_range(0, duration)
        if not self._seeking:
            self.seek_bar.set_value(position)

    def on_state_changed(self, player, state):
        """Show the state of the player."""
        if state == Gst.State.PLAYING:
            icon_name = get_icon_name([
                'media-playback-pause-symbolic', 'media-playback-pause'])
        else:
            icon_name = get_icon_name([
                'media-playback-start-symbolic', 'media-playback-start'])

        self.toggle_button.set_image(Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.BUTTON))

    def on_seeking_start(self, seek_bar, event):
        """Called when the user clicks on the seek bar."""
        if event.button == 1:
            self._seeking = True

    def on_seeking_end(self, seek_bar, event):
        """Called when the user release the seek bar."""
        if event.button == 1:
            self._seeking = False
            self.player.seek(seek_bar.get_value())

    def on_speed_scale_changed(self, scale, scroll, value):
        """Called when the user changes the value of the speed scale."""
        # The value in the parameters may be bigger than the maximum value of
        # the scale. Overwrite it with the actual value of the scale.
        value = scale.get_value()

        # Convert the value of the scale to a speed. This makes the scale
        # logarithmic (adding 12.5 to the value of the scale multiplies the
        # speed by two).
        speed = 6.25 * np.exp(np.log(2) * value / 12.5)

        self.speed_label.set_text('{:.1f} %'.format(speed))

    def on_speed_scale_released(self, scale, event):
        """Called when the user releases the speed scale."""
        # The value in the parameters may be bigger than the maximum value of
        # the scale. Overwrite it with the actual value of the scale.
        value = scale.get_value()

        # Convert the value of the scale to a speed. This makes the scale
        # logarithmic (adding 12.5 to the value of the scale multiplies the
        # speed by two).
        speed = 6.25 * np.exp(np.log(2) * value / 12.5)

        self.speed_label.set_text('{:.1f} %'.format(speed))
        self.player.set_speed(speed / 100)

    def on_toggle(self, _):
        """Method called when the play/pause button is clicked."""
        self.player.toggle()


def main():
    """Run audiotsmgtk."""
    if len(sys.argv) <= 1:
        print('usage: audiotsmgtk.py <filename>')
        return

    window = MainWindow()
    window.open_file(sys.argv[1])

    window.show_all()

    GObject.threads_init()
    Gtk.main()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Tests for the audiotsm.phasevocoder package.
"""

import pytest
import numpy as np
from numpy.testing import assert_array_equal

from audiotsm.phasevocoder import find_peaks, get_closest_peaks


@pytest.mark.parametrize("amplitude, peaks", [
    ([], []),
    ([0], [True]),
    ([1], [True]),
    ([1, 0, 0, 1, 0, 0, 0, 1, 0],
     [True, False, False, True, False, False, False, True, False]),
    ([0, 1, 2, 3, 1, 2, 4, 5, -1, 6, 0],
     [False, False, False, True, False, False, False, False, False, True,
      False]),
])
def test_find_peaks(amplitude, peaks):
    """Run tests for the find_peaks function."""
    actual_peaks = find_peaks(np.array(amplitude))
    assert_array_equal(actual_peaks, np.array(peaks, dtype=bool))


@pytest.mark.parametrize("peaks, closest_peak", [
    ([], []),
    ([True], [0]),
    ([True], [0]),
    ([True, False, False, True, False, False, False, True, False],
     [0, 0, 3, 3, 3, 3, 7, 7, 7]),
    ([True, True, True, True, True, True, True, True, True],
     [0, 1, 2, 3, 4, 5, 6, 7, 8]),
    ([False, False, False, True, False, False, False, False, False, True,
      False],
     [3, 3, 3, 3, 3, 3, 3, 9, 9, 9, 9]),
])
def test_get_closest_peaks(peaks, closest_peak):
    """Run tests for the get_closest_peaks."""
    actual_closest_peak = get_closest_peaks(np.array(peaks))
    assert_array_equal(actual_closest_peak, np.array(closest_peak, dtype=int))
